
ATTEMPTS = 1
SJUMP = 1000

import m5
from m5.objects import *
import os, optparse, sys
m5.AddToPath('../common')
import Simulation
from Caches import *
import pickle, random


# Get paths we might need.  It's expected this file is in m5/configs/example.
config_path = os.path.dirname(os.path.abspath(__file__))
config_root = os.path.dirname(config_path)
m5_root = os.path.dirname(config_root)

parser = optparse.OptionParser()

# Benchmark options
parser.add_option("-c", "--cmd",
                  default='lump.dat',
                  help="The parameter file (pickle).")
parser.add_option("-o", "--options", default="",
                  help="Not used.")
parser.add_option("-i", "--input", default="",
                  help="Not used.")

execfile(os.path.join(config_root, "common", "Options.py"))

(options, args) = parser.parse_args()

if args:
    print "Error: script doesn't take any positional arguments"
    sys.exit(1)

vt_parameters = pickle.load(file(options.cmd))
options.cmd = vt_parameters[ 'cmd' ]
meta_in = vt_parameters[ 'meta_in' ]
meta_out = vt_parameters[ 'meta_out' ]
max_length = vt_parameters[ 'max_length' ]
warnings_out = vt_parameters[ 'warnings_out' ]
start_pc = vt_parameters[ 'start_pc' ]
capture_timings = vt_parameters[ 'capture_timings' ]
runtime = vt_parameters[ 'runtime' ]


process = LiveProcess()
process.executable = options.cmd
process.cmd = [options.cmd] + options.options.split()
if options.input != "":
    process.input = options.input

if options.detailed:
    #check for SMT workload
    workloads = options.cmd.split(';')
    if len(workloads) > 1:
        process = []
        smt_idx = 0
        inputs = []

        if options.input != "":
            inputs = options.input.split(';')

        for wrkld in workloads:
            smt_process = LiveProcess()
            smt_process.executable = wrkld
            smt_process.cmd = wrkld + " " + options.options
            if inputs and inputs[smt_idx]:
                smt_process.input = inputs[smt_idx]
            process += [smt_process, ]
            smt_idx += 1

(CPUClass, test_mem_mode, FutureClass) = Simulation.setCPUClass(options)

CPUClass.clock = '1GHz'

np = options.num_cpus

system = System(cpu = [CPUClass(cpu_id=i) for i in xrange(np)],
                physmem = PhysicalMemory(range=AddrRange("512MB")),
                membus = Bus(), mem_mode = test_mem_mode)

system.physmem.port = system.membus.port

if options.l2cache:
    system.l2 = L2Cache(size='2MB')
    system.tol2bus = Bus()
    system.l2.cpu_side = system.tol2bus.port
    system.l2.mem_side = system.membus.port

for i in xrange(np):
    if options.caches:
        system.cpu[i].addPrivateSplitL1Caches(L1Cache(size = '64kB'),
                                              L1Cache(size = '64kB'))
    if options.l2cache:
        system.cpu[i].connectMemPorts(system.tol2bus)
    else:
        system.cpu[i].connectMemPorts(system.membus)
    system.cpu[i].workload = process

    if options.fastmem:
        system.cpu[0].physmem_port = system.physmem.port

root = Root(system = system)



vt = VirtualTrace()
vt.loadMetaInstructions(meta_in)
m5.instantiate(root)

def Spin():
    val = -1
    watchdog = 10000
    while (( val < 0 ) and ( watchdog >= 0 )):
        m5.simulate(SJUMP)
        val = vt.getMeasurement()
        watchdog -= 1

    if ( watchdog < 0 ):
        print 'XXX timeout'
        sys.exit(1)

    return val


def Capture_All_Timings():
    check_measurement = dict()
    warnings = file(warnings_out, 'wt')
    warnings.flush()
    meta_inst = []
    for line in file(meta_in):
        if ( line.startswith('STA ') ):
            addr = int(line[ 4: ], 16)
            meta_inst.append(addr)

    #meta_inst = [ 
    #    0x120000414, 0x120000414, 0x120000414, 0x120000414,
    #    0x120063df4, 0x120009fb8 ]

    random.seed(1)

    vt.setMeasureMode(True)

    valid_count = total_count = 0
    for y in xrange(ATTEMPTS):
        for x in xrange(max_length):
            random.shuffle(meta_inst)
            for start in meta_inst:
                print 'XXX Measure %x (attempt %u length %u)' % (start, y, x)
                vt.selectStart(start, x)
                val = Spin()

                key = (start, x)
                if ( y > 0 ): # ( ATTEMPTS - 2 ) ):
                    if ( check_measurement[ key ] != val ):
                        warnings.write('For start %x length %u, '
                        'measurement changed from %u to %u at attempt %u.\n' %
                                (start, x, 
                                    check_measurement[ key ], val, y))
                        warnings.flush()
                    else:
                        valid_count += 1

                    total_count += 1

                check_measurement[ key ] = val

        #print 'Checkpoint save/restore'
        #m5.checkpoint(root, "cp1")
        #m5.restoreCheckpoint(root, "cp1")
    
    warnings.write('Unchanged measurements in %u of %u cases.\n' % (
            valid_count, total_count))
    warnings.close()
    vt.saveMetaInstructions(meta_out)

def Fast_Warmup():
    print 'XXX Warmup'
    ok = vt.setMeasureMode(True)
    if not ok:
        return

    for i in range(2):
        vt.selectStart(start_pc, ( max_length / 2 ) + 1)
        Spin()
        vt.selectStart(start_pc + 4, ( max_length / 2 ) + 1)
        Spin()

def Exec_Mode(runtime):
    print 'XXX Going to execution mode'
    vt.setMeasureMode(False)
    vt.selectStart(start_pc)
    start = vt.getCurrentTime()
    m5.simulate(int(runtime))
    stop = vt.getCurrentTime()
    print 'start:', start
    print 'stop:', stop

    vt.saveMetaInstructions(meta_out)

m5.setVirtualTrace(system.cpu[0], vt)

if ( capture_timings ):
    Capture_All_Timings()
else:
    Fast_Warmup()

if ( runtime > 0 ):
    Exec_Mode(runtime)
    print 'magic number', vt.getBenchmarkTime()
    print vt.getConfiguration()


