
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <list>
#include "ftr.hh"
#include "unix.hh"


#define NUM_REGS        16

#define M5_CMD          "m5/build/ARM_SE/m5.opt "
#define QEMU_FIFO       "/tmp/qemubin.log"
#define M5_FIFO         "/tmp/m5.log"

bool Regression_QEMU_M5 ( Symbol_Table * symtab ,
                FILE * qemu_trace , FILE * m5_trace )
{
    struct FastTraceRecord      ftrp [ 1 ] ;
    struct FastTraceRecord *    ftr0 = & ftrp [ 0 ] ;
    ulongint                    count = 0 ;
    ulongint                    pc_history_size = 0 ;
    std::list<ulongint>         pc_history ;
    std::list<ulongint>::iterator i ;
    bool                        invalid , retry , src ;
    ulongint                    qemu_pc = 0 ;
    ulongint                    previous_pc = 0 ;
    ulongint                    cdown = 0 ;
    uint32_t                    qemu_register [ NUM_REGS ] ;
    time_t                      total_time , start_time = time ( NULL ) ;
    bool                        reached_main = false , stop = false ;
    ulongint                    j , main_end , main_start , main_size ;
    int                         last_out_reg_number = 0 ;
    ulongint                    last_out_value = 0 ;

    memset ( qemu_register , 0 , sizeof ( qemu_register ) ) ;
    src = symtab -> Get_Record ( "main" , main_start , main_size ) ;
    if ( ! src )
    {
        fprintf ( stderr , "Unable to look up address of 'main'.\n" ) ;
        return false ;
    }
    main_end = main_start + main_size ;
    fprintf ( stderr , "main at 0x%"FX" to 0x%"FX"\n" , main_start , main_end ) ;

    while ( 1 )
    {
        if ( cdown == 0 )
        {
            double rate = 0.0 ;

            total_time = time ( NULL ) - start_time ;
            if ( total_time != 0 )
            {
                rate = ( (double) count / (double) total_time ) / 1000.0 ;
            }
            fprintf ( stderr , 
                "%"FU" instructions, %1.0fk per second, pc 0x%"FX"\n" , 
                        count , rate , qemu_pc ) ;
            cdown = 100000 - 1 ;
        } else {
            cdown -- ;
        }
        count ++ ;

        /* Get record from QEMU */
        invalid = ( fread ( qemu_register , 
                    NUM_REGS * 4 , 1 , qemu_trace ) != 1 ) ;

        if ( invalid )
        {
            fprintf ( stderr , "\nInstruction %"FU" @ %"FU": "
                "EOF on data from QEMU.\n" , count , ftr0 -> when ) ;
            fprintf ( stderr , "Last instruction: 0x%"FX"\n" , 
                    ftr0 -> pc_value ) ;
            return true ;
        }
        /* Interpret QEMU record */
        qemu_pc = qemu_register [ NUM_REGS - 1 ] ;

        retry = true ;
        while ( retry )
        {
            /* Get record from M5 */
            invalid = true ;
            while ( invalid && ( fread ( ftr0 , 
                sizeof ( struct FastTraceRecord ) , 1 , m5_trace ) == 1 ))
            {
                invalid = (( ftr0 -> optype == 'T' )
                    || ( ftr0 -> pc_value == previous_pc )) ;
            }

            if ( invalid )
            {
                fprintf ( stderr , "\nInstruction %"FU" @ %"FU": "
                    "EOF on data from M5.\n" , count , ftr0 -> when ) ;
                fprintf ( stderr , "Last instruction: 0x%"FX"\n" , 
                        ftr0 -> pc_value ) ;
                retry = false ;
                stop = true ;
#if 0
            } else if (( ftr0 -> pc_value < qemu_pc )
            && ( ftr0 -> pc_value == ( previous_pc + 4 )))
            {
                /* This could be a non-executed sequence of instructions */
                retry = false ;
                stop = true ;
#endif
            } else if ( ftr0 -> pc_value < qemu_pc )
            { 
                /* Running behind? */
                retry = false ;
                stop = true ;
            } else if ( ftr0 -> pc_value > qemu_pc )
            { 
                /* Running ahead? */
                retry = false ;
                stop = true ;
            } else if ( ftr0 -> pc_value == qemu_pc )
            { 
                /* Perfect */
                retry = false ;
            }

            /* Log historic data */
            previous_pc = ftr0 -> pc_value ;
            pc_history . push_back ( ftr0 -> pc_value |
                        ( retry ? 1 : 0 )) ;
            if ( pc_history_size >= 100 )
            {
                pc_history . pop_front () ;
            } else {
                pc_history_size ++ ;
            }
        }



        /* Compare records */
        if (( last_out_reg_number < NUM_REGS )
        && ( last_out_value < ( (ulongint) 1 << (ulongint) 32 ))
        && ( qemu_register[ last_out_reg_number ] !=
                    last_out_value ))
        {
            if ( reached_main )
            {
                fprintf ( stderr , "Stop! " ) ;
                stop = true ;
            } else {
                fprintf ( stderr , "%08"FX": Before main: " , qemu_pc ) ;
            }
            fprintf ( stderr , "QEMU and M5 registers do not match\n" ) ;
            fprintf ( stderr , "pc 0x%"FX" is " , ftr0 -> pc_value ) ;
            symtab -> Lookup ( stderr , ftr0 -> pc_value ) ;
            fprintf ( stderr , "Output register is r%d\n"
                    "QEMU value 0x%08"FX"\n" 
                    "M5 value 0x%08"FX"\n" , 
                            (int) last_out_reg_number ,
                            (ulongint) qemu_register [ last_out_reg_number ] ,
                            (ulongint) last_out_value ) ;
            for ( j = 0 ; j < NUM_REGS ; j ++ )
            {
                fprintf ( stderr , "  r%-2d = 0x%08"FX"\n" ,
                        (int) j , (ulongint) qemu_register [ j ] ) ;
            }
            fprintf ( stderr , "\n" ) ;
        } else if (( last_out_reg_number < NUM_REGS )
        && ( last_out_value < ( (ulongint) 1 << (ulongint) 32 )))
        {
            //fprintf ( stderr , "%08"FX": Match r%d value %08"FX"\n" ,
            //                qemu_pc ,
            //                (int) last_out_reg_number ,
            //                qemu_register [ last_out_reg_number ] ) ;
        } else {
            //fprintf ( stderr , "%08"FX": Skip\n" , qemu_pc ) ;
        }

        last_out_value = ftr0 -> out_value ;
        last_out_reg_number = ftr0 -> out_reg_number ;

        if ( ftr0 -> pc_value != qemu_pc )
        {
            fprintf ( stderr , "Stop! QEMU and M5 PCs do not match\n" ) ;
            stop = true ;
        }

        if ( stop )
        {
            fprintf ( stderr , "History:" ) ;

            for ( i = pc_history . begin () ; 
                    i != pc_history . end () ; i ++ )
            {
                ulongint pc = (* i) & ~1 ;
                fprintf ( stderr , "\n    %c0x%08"FX" is " , 
                        ((* i) & 1 ) ? '*' : ' ' , pc ) ;
                symtab -> Lookup ( stderr , pc ) ;
            }
            fprintf ( stderr , "\nInstruction %"FU" @ %"FU
                    ": M5 at pc 0x%"FX", "
                    "QEMU at pc 0x%"FX"." , count , ftr0 -> when ,
                    ftr0 -> pc_value , qemu_pc ) ;
            fprintf ( stderr , "\npc 0x%"FX" is " , ftr0 -> pc_value ) ;
            symtab -> Lookup ( stderr , ftr0 -> pc_value ) ;
            fprintf ( stderr , "\npc 0x%"FX" is " , qemu_pc ) ;
            symtab -> Lookup ( stderr , qemu_pc ) ;
            fprintf ( stderr , "\n\n" ) ;
            return false ;
        }

        if (( ! reached_main )
        && ( main_start <= qemu_pc )
        && ( qemu_pc < main_end ))
        { 
            reached_main = true ;
        }
    }
}

bool Harness_QEMU_M5 ( std::string argv , std::string program )
{
    Symbol_Table * symtab = Do_Read_Symbols ( program ,
            "./bin/arm-9tdmi-linux-gnu-nm" ) ;

    if ( symtab == 0L )
    {
        fprintf ( stderr , "Error reading symbols from program %s\n" ,
                program . c_str () ) ;
        return false ;
    }

    unlink ( QEMU_FIFO ) ;
    mkfifo ( QEMU_FIFO , 0600 ) ;
    setenv ( "QEMU_BIN_TRACE_OUT" , QEMU_FIFO , 1 ) ;
    unlink ( M5_FIFO ) ;
    mkfifo ( M5_FIFO , 0600 ) ;

    char cmd [ BUFSIZ ] ;
    cmd [ BUFSIZ - 1 ] = '\0' ;
    snprintf ( cmd , BUFSIZ - 1 ,
            M5_CMD " --trace-flags=Exec,ExecFTR "
            "--trace-file " M5_FIFO " m5/configs/example/se.py "
            "%s -c %s 1>&2" , argv . c_str () , program . c_str () ) ;
    fprintf ( stderr , "M5: %s\n" , cmd ) ;

    if ( popen ( cmd , "r" ) == 0L )
    {
        perror ( "Unable to start M5." ) ;
        return false ;
    }

    snprintf ( cmd , BUFSIZ - 1 ,
            "qemu-0.10.5/arm-linux-user/qemu-arm %s" ,
            program . c_str () ) ;
    fprintf ( stderr , "QEMU: %s\n" , cmd ) ;

    if ( popen ( cmd , "r" ) == 0L )
    {
        perror ( "Unable to start M5." ) ;
        return false ;
    }

    FILE * m5_trace = fopen ( M5_FIFO , "rb" ) ;
    if ( m5_trace == 0L )
    {
        perror ( "Unable to open M5 FIFO " M5_FIFO ) ;
        return false ;
    }
    FILE * qemu_trace = fopen ( QEMU_FIFO , "rt" ) ;
    if ( qemu_trace == 0L )
    {
        perror ( "Unable to open QEMU FIFO " QEMU_FIFO ) ;
        return false ;
    }
    fprintf ( stderr , "\nFIFOs open, starting test.\n" ) ;
    return Regression_QEMU_M5 ( symtab , qemu_trace , m5_trace ) ;
}

int main ( int argc , char ** argv )
{
    std::string cc_argv ;
    int i ;

    for ( i = 1 ; i < argc ; i ++ )
    {
        cc_argv += " " ;
        cc_argv += argv [ i ] ;
    }
    if ( Harness_QEMU_M5 ( cc_argv , "tc/testcase" )
    && Harness_QEMU_M5 ( cc_argv , "linuxdoom-1.10/linuxxdoom" ))
    {
        return 0 ;
    } else {
        return 1 ;
    }
}


